from zrb.helper.typing import Optional
from zrb.helper.typecheck import typechecked
import os


@typechecked
class Env():
    '''
    Task Environment definition
    '''

    def __init__(
        self,
        name: str,
        os_name: Optional[str] = None,
        default: str = '',
        renderable: bool = True,
    ):
        self.name: str = name
        self.os_name: str = os_name if os_name is not None else name
        self.default: str = default
        self.renderable: bool = renderable

    def get(self, prefix: str = '') -> str:
        '''
        Return environment value.
        You can use prefix to distinguish 'DEV', 'PROD'

        Example:
        ```python
        os.environ['DEV_SERVER'] = 'localhost'
        os.environ['PROD_SERVER'] = 'example.com'

        env = Env(name='HOST', os_name='SERVER', default='0.0.0.0')

        print(env.get('DEV'))   # will show 'localhost'
        print(env.get('PROD'))  # will show 'example.com'
        print(env.get('STAG'))  # will show '0.0.0.0'
        ```
        '''
        if self.os_name == '':
            return self.default
        prefixed_name = self._get_prefixed_name(self.os_name, prefix)
        if prefixed_name in os.environ and os.environ[prefixed_name] != '':
            return os.environ[prefixed_name]
        if self.os_name in os.environ and os.environ[self.os_name] != '':
            return os.environ[self.os_name]
        return self.default

    def _get_prefixed_name(self, name: str, prefix: str):
        if prefix is None or prefix == '':
            return name
        return prefix + '_' + name

    def __repr__(self) -> str:
        name = self.name
        os_name = self.os_name
        default = self.default
        return f'<Env name={name} os_name={os_name} default={default}>'



from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Optional
import libcst as cst


@typechecked
class AddPropertyTransformer(cst.CSTTransformer):
    def __init__(
        self,
        class_name: str,
        property_name_node: cst.Name,
        property_type_node: cst.Annotation,
        property_value_node: Optional[cst.BaseExpression]
    ):
        self.class_name = class_name
        self.property_name_node = property_name_node
        self.property_type_node = property_type_node
        self.property_value_node = property_value_node
        super().__init__()

    def leave_ClassDef(
        self, original_node: cst.ClassDef, updated_node: cst.ClassDef
    ) -> cst.ClassDef:
        if updated_node.name.value != self.class_name:
            return updated_node
        class_body_statements = []
        if not isinstance(original_node.body, cst.EmptyLine):
            for stmt in original_node.body.body:
                if (
                    hasattr(stmt, 'body') and
                    len(stmt.body) > 0 and
                    isinstance(stmt.body[0], cst.Pass)
                ):
                    continue
                # Add all old class body except any `pass` statement
                class_body_statements.append(stmt)
        # Add new property to the class
        class_body_statements.append(
            cst.SimpleStatementLine(
                body=[
                    cst.AnnAssign(
                        target=self.property_name_node,
                        annotation=self.property_type_node,
                        value=self.property_value_node,
                    )
                ]
            )
        )
        new_body = cst.IndentedBlock(
            body=class_body_statements,
            indent=updated_node.body.indent,
        )
        return updated_node.with_changes(body=new_body)


@typechecked
def add_property_to_class(
    code: str,
    class_name: str,
    property_name: str,
    property_type: str,
    property_value: Optional[str] = None
) -> str:
    module = cst.parse_module(code)
    property_name_node = cst.Name(value=property_name)
    property_type_node = cst.Annotation(cst.Name(value=property_type))
    property_value_node = _get_property_value_node(property_value)
    transformed_module = module.visit(
        AddPropertyTransformer(
            class_name=class_name,
            property_name_node=property_name_node,
            property_type_node=property_type_node,
            property_value_node=property_value_node
        )
    )
    return transformed_module.code


def _get_property_value_node(
    property_value: Optional[str]
) -> Optional[cst.BaseExpression]:
    if property_value is None:
        return None
    return cst.parse_expression(property_value)



from zrb import (
    runner, CmdTask, ResourceMaker, DockerComposeTask, FlowTask,
    Env, BoolInput, StrInput, HTTPChecker, Group
)
import os
import sys
import tomli

CURRENT_DIR = os.path.dirname(__file__)
PLAYGROUND_DIR = os.path.join(CURRENT_DIR, 'playground')
IS_PLAYGROUND_EXIST = os.path.isdir(PLAYGROUND_DIR)

if IS_PLAYGROUND_EXIST:
    sys.path.append(PLAYGROUND_DIR)
    from playground import zrb_init
    assert zrb_init


with open(os.path.join(CURRENT_DIR, 'pyproject.toml'), 'rb') as f:
    toml_dict = tomli.load(f)
    VERSION = toml_dict['project']['version']


###############################################################################
# Group Definitions
###############################################################################

playground_group = Group(
    name='playground', description='Playground related tasks'
)

###############################################################################
# Input Definitions
###############################################################################

zrb_image_name_input = StrInput(
    name='zrb-image-name',
    description='Zrb image name',
    prompt='Zrb image name',
    default=f'docker.io/stalchmst/zrb:{VERSION}'
)

zrb_latest_image_name_input = StrInput(
    name='zrb-latest-image-name',
    description='Zrb latest image name',
    prompt='Zrb latest image name',
    default='docker.io/stalchmst/zrb:latest'
)

build_zrb_input = BoolInput(name='build-zrb', default=True)
install_symlink_input = BoolInput(name='install-symlink', default=True)
create_playground_input = BoolInput(
    name='create-playground', default=not IS_PLAYGROUND_EXIST
)

###############################################################################
# Env Definitions
###############################################################################

zrb_image_env = Env(
    name='ZRB_IMAGE',
    os_name='',
    default='{{input.zrb_image_name}}'
)

zrb_latest_image_env = Env(
    name='ZRB_IMAGE',
    os_name='',
    default='{{input.zrb_latest_image_name}}'
)

###############################################################################
# Task Definitions
###############################################################################

build = CmdTask(
    name='build',
    description='Build Zrb',
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Build zrb distribution"',
        f'rm -Rf {CURRENT_DIR}/dist',
        'git add . -A',
        'flit build',
    ],
)
skippable_build: CmdTask = build.copy()
skippable_build.add_inputs(build_zrb_input)
skippable_build.set_skip_execution('{{ not input.build_zrb}}')
runner.register(build)

publish_pip = CmdTask(
    name='publish-pip',
    description='Publish zrb to pypi',
    upstreams=[build],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Publish zrb to pypi"',
        'flit publish --repository pypi',
    ]
)
runner.register(publish_pip)

publish_pip_test = CmdTask(
    name='publish-pip-test',
    description='Publish zrb to testpypi',
    upstreams=[build],
    cwd=CURRENT_DIR,
    cmd=[
        'set -e',
        'echo "🤖 Publish zrb to testpypi"',
        'flit publish --repository testpypi',
    ]
)
runner.register(publish_pip_test)

prepare_docker = ResourceMaker(
    name='prepare-docker',
    description='Create docker directory',
    template_path=f'{CURRENT_DIR}/docker-template',
    destination_path=f'{CURRENT_DIR}/.docker-dir',
    replacements={
        'ZRB_VERSION': VERSION
    }
)
runner.register(prepare_docker)

check_pip = HTTPChecker(
    name='check-pip',
    is_https=True,
    host='pypi.org',
    url=f'pypi/zrb/{VERSION}/json',
    port=443
)

build_image = DockerComposeTask(
    name='build-image',
    description='Build docker image',
    upstreams=[
        prepare_docker,
        check_pip,
    ],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='build',
    compose_args=['zrb']
)
runner.register(build_image)

build_latest_image = DockerComposeTask(
    name='build-latest-image',
    description='Build docker image',
    upstreams=[
        prepare_docker,
        check_pip,
        build_image,
    ],
    inputs=[zrb_latest_image_name_input],
    envs=[zrb_latest_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='build',
    compose_args=['zrb']
)
runner.register(build_latest_image)

stop_container = DockerComposeTask(
    name='stop-container',
    description='remove docker container',
    upstreams=[prepare_docker],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='down'
)
runner.register(stop_container)

start_container = DockerComposeTask(
    name='start-container',
    description='Run docker container',
    upstreams=[
        build_image,
        stop_container
    ],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='up',
    compose_flags=['-d']
)
runner.register(start_container)

push_image = DockerComposeTask(
    name='push-image',
    description='Push docker image',
    upstreams=[build_image],
    inputs=[zrb_image_name_input],
    envs=[zrb_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='push',
    compose_args=['zrb']
)
runner.register(push_image)

push_latest_image = DockerComposeTask(
    name='push-latest-image',
    description='Push docker image',
    upstreams=[
        build_latest_image,
        push_image,
    ],
    inputs=[zrb_latest_image_name_input],
    envs=[zrb_latest_image_env],
    cwd=f'{CURRENT_DIR}/.docker-dir',
    compose_cmd='push',
    compose_args=['zrb']
)
runner.register(push_latest_image)

publish = FlowTask(
    name='publish',
    description='Publish new version',
    steps=[
        publish_pip,
        push_latest_image,
    ]
)
runner.register(publish)

install_symlink = CmdTask(
    name='install-symlink',
    description='Install Zrb as symlink',
    upstreams=[skippable_build],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Install zrb"',
        'flit install --symlink',
    ]
)
skippable_install_symlink: CmdTask = install_symlink.copy()
skippable_install_symlink.add_inputs(
    build_zrb_input,
    install_symlink_input
)
skippable_install_symlink.set_skip_execution('{{ not input.install_symlink}}')
runner.register(install_symlink)

test = CmdTask(
    name='test',
    description='Run zrb test',
    inputs=[
        StrInput(
            name='test',
            shortcut='t',
            description='Specific test case (i.e., test/file.py::test_name)',
            prompt='Test (i.e., test/file.py::test_name)',
            default=''
        ),
    ],
    upstreams=[install_symlink],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Perform test"',
        'pytest --ignore-glob="**/template/**/test" --ignore=playground --cov=zrb --cov-report=html --cov-report=term --cov-report=term-missing {{input.test}}'  # noqa
    ],
    retry=0,
    checking_interval=1
)
runner.register(test)

serve_test = CmdTask(
    name='serve-test',
    description='Serve zrb test result',
    inputs=[
        StrInput(
            name='port',
            shortcut='p',
            description='Port to serve coverage result',
            prompt='Serve coverage on port',
            default='9000'
        )
    ],
    upstreams=[test],
    cmd=[
        'set -e',
        f'cd {CURRENT_DIR}',
        'echo "🤖 Serve coverage report"',
        f'python -m http.server {{input.port}} --directory {CURRENT_DIR}/htmlcov',  # noqa
    ],
    checkers=[
        HTTPChecker(port='{{input.port}}')
    ],
    retry=0,
    checking_interval=0.3
)
runner.register(serve_test)

create_playground = CmdTask(
    name='create',
    description='Create playground',
    group=playground_group,
    upstreams=[skippable_install_symlink],
    cwd=CURRENT_DIR,
    cmd_path=os.path.join(CURRENT_DIR, 'playground-create.sh'),
    retry=0,
    preexec_fn=None
)
skippable_create_playground: CmdTask = create_playground.copy()
skippable_create_playground.add_inputs(
    build_zrb_input,
    install_symlink_input,
    create_playground_input
)
skippable_create_playground.set_skip_execution('{{ not input.create_playground}}') # noqa
runner.register(create_playground)

test_fastapp_playground = CmdTask(
    name='test-fastapp',
    description='Test Fastapp',
    group=playground_group,
    upstreams=[skippable_create_playground],
    cwd=CURRENT_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'playground-init.sh'),
        os.path.join(CURRENT_DIR, 'playground-test-fastapp.sh')
    ],
    retry=0,
    preexec_fn=None
)
runner.register(test_fastapp_playground)

test_install_playground_symlink = CmdTask(
    name='test-install-symlink',
    description='Test installing symlink',
    group=playground_group,
    upstreams=[skippable_create_playground],
    cwd=CURRENT_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, 'playground-init.sh'),
        os.path.join(CURRENT_DIR, 'playground-test-install-symlink.sh')
    ],
    retry=0,
    preexec_fn=None
)
runner.register(test_install_playground_symlink)

test_playground = CmdTask(
    name='test',
    description='Test playground',
    group=playground_group,
    upstreams=[
        test_fastapp_playground,
        test_install_playground_symlink
    ],
    cmd='echo Test performed',
    retry=0
)
runner.register(test_playground)



import logging

LOGGING_LEVEL_MAP = {
    'critical': logging.CRITICAL,
    'fatal': logging.FATAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'warn': logging.WARN,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'notset': logging.NOTSET,
}


def str_to_boolean(value: str):
    if value.lower() in ('0', 'false', 'no', 'n'):
        return False
    if value.lower() in ('1', 'true', 'yes', 'y'):
        return True
    raise Exception(f'Cannot convert to boolean: "{value}"')


def str_to_logging_level(logging_level_str: str) -> int:
    lower_logging_level_str = logging_level_str.lower()
    if lower_logging_level_str in LOGGING_LEVEL_MAP:
        return LOGGING_LEVEL_MAP[lower_logging_level_str]
    return logging.WARNING



from typing import TypeVar
from zrb.helper.string.constant import FALSE_STRS
from beartype import beartype
import os

enable_type_checking_str = os.getenv('ZRB_ENABLE_TYPE_CHECKING', '1').lower()
enable_type_checking = enable_type_checking_str not in FALSE_STRS

T = TypeVar('T')


def typechecked(anything: T) -> T:
    if enable_type_checking:
        return beartype(anything)
    return anything



from logging import Logger
from core.messagebus import Publisher, Consumer
from core.rpc import Caller


def register_event(
    logger: Logger,
    consumer: Consumer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Register event handlers for "snake_zrb_module_name"')



from zrb.helper.typing import (
    Any, Callable, Iterable, Mapping, Optional, Union, TypeVar
)
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.task.cmd_task import CmdTask, CmdVal

import copy
import os
import pathlib


TSingleBaseRemoteCmdTask = TypeVar(
    'TSingleBaseRemoteCmdTask', bound='SingleBaseRemoteCmdTask'
)
TBaseRemoteCmdTask = TypeVar(
    'TBaseRemoteCmdTask', bound='BaseRemoteCmdTask'
)


@typechecked
class RemoteConfig:
    def __init__(
        self,
        host: str,
        user: str = '',
        password: str = '',
        ssh_key: str = '',
        port: int = 22
    ):
        self.host = host
        self.user = user
        self.password = password
        self.ssh_key = ssh_key
        self.port = port


@typechecked
class SingleBaseRemoteCmdTask(CmdTask):
    def __init__(
        self,
        name: str,
        remote_config: RemoteConfig,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        executable: Optional[str] = None,
        pre_cmd: CmdVal = '',
        pre_cmd_path: CmdVal = '',
        cmd: CmdVal = '',
        cmd_path: CmdVal = '',
        post_cmd: CmdVal = '',
        post_cmd_path: CmdVal = '',
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        skip_execution: Union[bool, str, Callable[..., bool]] = False,
        return_upstream_result: bool = False
    ):
        CmdTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            cmd=cmd,
            cmd_path=cmd_path,
            cwd=cwd,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            skip_execution=skip_execution,
            return_upstream_result=return_upstream_result
        )
        self._pre_cmd = pre_cmd
        self._pre_cmd_path = pre_cmd_path
        self._post_cmd = post_cmd
        self._post_cmd_path = post_cmd_path
        self._remote_config = remote_config

    def copy(self) -> TSingleBaseRemoteCmdTask:
        return copy.deepcopy(self)

    def _get_shell_env_map(self) -> Mapping[str, Any]:
        env_map = super()._get_shell_env_map()
        env_map['_CONFIG_HOST'] = self.render_str(self._remote_config.host)
        env_map['_CONFIG_PORT'] = str(self.render_int(
            self._remote_config.port)
        )
        env_map['_CONFIG_SSH_KEY'] = self.render_str(
            self._remote_config.ssh_key
        )
        env_map['_CONFIG_USER'] = self.render_str(self._remote_config.user)
        env_map['_CONFIG_PASSWORD'] = self.render_str(
            self._remote_config.password
        )
        return env_map

    def _get_cmd_str(self, *args: Any, **kwargs: Any) -> str:
        cmd_str = '\n'.join([
            self._create_cmd_str(
                self._pre_cmd_path, self._pre_cmd, *args, **kwargs
            ),
            super()._get_cmd_str(*args, **kwargs),
            self._create_cmd_str(
                self._post_cmd_path, self._post_cmd, *args, **kwargs
            ),
        ])
        return cmd_str


@typechecked
class BaseRemoteCmdTask(BaseTask):
    def __init__(
        self,
        name: str,
        remote_configs: Iterable[RemoteConfig],
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        executable: Optional[str] = None,
        pre_cmd: CmdVal = '',
        pre_cmd_path: CmdVal = '',
        cmd: CmdVal = '',
        cmd_path: CmdVal = '',
        post_cmd: CmdVal = '',
        post_cmd_path: CmdVal = '',
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        skip_execution: Union[bool, str, Callable[..., bool]] = False,
        return_upstream_result: bool = False
    ):
        sub_tasks = [
            SingleBaseRemoteCmdTask(
                name=f'{name}-{remote_config.host}',
                remote_config=remote_config,
                inputs=inputs,
                envs=envs,
                env_files=env_files,
                description=description,
                executable=executable,
                pre_cmd=pre_cmd,
                pre_cmd_path=pre_cmd_path,
                cmd=cmd,
                cmd_path=cmd_path,
                post_cmd=post_cmd,
                post_cmd_path=post_cmd_path,
                cwd=cwd,
                upstreams=upstreams,
                checkers=checkers,
                checking_interval=checking_interval,
                retry=retry,
                retry_interval=retry_interval,
                max_output_line=max_output_line,
                max_error_line=max_error_line,
                preexec_fn=preexec_fn,
                skip_execution=skip_execution,
                return_upstream_result=return_upstream_result
            )
            for remote_config in list(remote_configs)
        ]
        BaseTask.__init__(
            self,
            name=name,
            icon=icon,
            color=color,
            group=group,
            description=description,
            upstreams=sub_tasks,
            retry=0,
            return_upstream_result=True
        )



from zrb.helper.typing import (
    Callable, Iterable, List, Optional, TypeVar, Union
)
from zrb.helper.typecheck import typechecked
from zrb.task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput


TFlowTask = TypeVar('TFlowTask', bound='FlowTask')


@typechecked
class FlowTask(BaseTask):

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: float = 0,
        retry: int = 2,
        retry_interval: float = 1,
        steps: List[Union[AnyTask, List[AnyTask]]] = [],
        skip_execution: Union[bool, str, Callable[..., bool]] = False,
        return_upstream_result: bool = False
    ):
        final_upstreams: List[AnyTask] = list(upstreams)
        inputs: List[AnyInput] = list(inputs)
        envs: List[Env] = list(envs)
        env_files: List[EnvFile] = list(env_files)
        for step in steps:
            tasks = self._step_to_tasks(step)
            new_upstreams = self._get_embeded_tasks(
                tasks=tasks,
                upstreams=final_upstreams,
                inputs=inputs,
                envs=envs,
                env_files=env_files
            )
            final_upstreams = new_upstreams
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=final_upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            skip_execution=skip_execution,
            return_upstream_result=return_upstream_result,
            run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
        )

    def copy(self) -> TFlowTask:
        return super().copy()

    def _step_to_tasks(
        self, node: Union[AnyTask, List[AnyTask]]
    ) -> List[AnyTask]:
        if isinstance(node, AnyTask):
            return [node]
        return node

    def _get_embeded_tasks(
        self,
        tasks: List[AnyTask],
        upstreams: List[AnyTask],
        inputs: List[AnyInput],
        envs: List[Env],
        env_files: List[EnvFile],
    ) -> List[AnyTask]:
        embeded_tasks: List[AnyTask] = []
        for task in tasks:
            embeded_task = task.copy()
            embeded_task.add_upstreams(*upstreams)
            embeded_task.add_envs(*envs)
            embeded_task.add_env_files(*env_files)
            embeded_task.add_inputs(*inputs)
            embeded_tasks.append(embeded_task)
        return embeded_tasks

    def __repr__(self) -> str:
        return f'<FlowTask name={self._name}>'



from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union, TypeVar
)
from zrb.helper.typecheck import typechecked
from zrb.task.cmd_task import CmdTask, CmdResult, CmdVal
from zrb.task.any_task import AnyTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.helper.accessories.name import get_random_name
from zrb.helper.string.conversion import to_cmd_name
from zrb.helper.string.modification import double_quote
from zrb.helper.docker_compose.file import (
    read_compose_file, write_compose_file
)
from zrb.helper.docker_compose.fetch_external_env import (
    fetch_compose_file_env_map
)

import os
import pathlib

CURRENT_DIR = os.path.dirname(__file__)
SHELL_SCRIPT_DIR = os.path.join(CURRENT_DIR, '..', 'shell-scripts')
TDockerComposeTask = TypeVar('TDockerComposeTask', bound='DockerComposeTask')

ensure_docker_is_installed = CmdTask(
    name='ensure-docker-is-installed',
    cmd_path=[
        os.path.join(SHELL_SCRIPT_DIR, '_common-util.sh'),
        os.path.join(SHELL_SCRIPT_DIR, 'ensure-docker-is-installed.sh')
    ],
    preexec_fn=None
)


@typechecked
class ServiceConfig():
    def __init__(self, envs: List[Env] = [], env_files: List[EnvFile] = []):
        self._envs = envs
        self._env_files = env_files

    def get_envs(self) -> List[Env]:
        return self._envs

    def get_env_files(self) -> List[EnvFile]:
        return self._env_files


@typechecked
class DockerComposeTask(CmdTask):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        executable: Optional[str] = None,
        compose_service_configs: Mapping[str, ServiceConfig] = {},
        compose_file: Optional[str] = None,
        compose_cmd: str = 'up',
        compose_options: Mapping[str, str] = {},
        compose_flags: Iterable[str] = [],
        compose_args: Iterable[str] = [],
        compose_env_prefix: str = '',
        setup_cmd: CmdVal = '',
        setup_cmd_path: CmdVal = '',
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: float = 0.1,
        retry: int = 2,
        retry_interval: float = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        skip_execution: Union[bool, str, Callable[..., bool]] = False,
        return_upstream_result: bool = False
    ):
        combined_env_files = list(env_files)
        CmdTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=combined_env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            cwd=cwd,
            upstreams=[ensure_docker_is_installed] + upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            skip_execution=skip_execution,
            return_upstream_result=return_upstream_result
        )
        self._setup_cmd = setup_cmd
        self._setup_cmd_path = setup_cmd_path
        self._compose_service_configs = compose_service_configs
        self._compose_cmd = compose_cmd
        self._compose_options = compose_options
        if '--f' in compose_options:
            raise ValueError('compose option cannot contains --f')
        if '--file' in compose_options:
            raise ValueError('compose option cannot contains --file')
        self._compose_flags = compose_flags
        self._compose_args = compose_args
        self._compose_env_prefix = compose_env_prefix
        self._compose_template_file = self._get_compose_template_file(
            compose_file
        )
        self._compose_runtime_file = self._get_compose_runtime_file(
            self._compose_template_file
        )
        # append services env and env_files to current task
        for _, service_config in self._compose_service_configs.items():
            self._env_files += service_config.get_env_files()
            self._envs += service_config.get_envs()
        self._add_compose_envs()

    def copy(self) -> TDockerComposeTask:
        return super().copy()

    async def run(self, *args, **kwargs: Any) -> CmdResult:
        self._generate_compose_runtime_file()
        try:
            result = await super().run(*args, **kwargs)
        finally:
            os.remove(self._compose_runtime_file)
        return result

    def _generate_compose_runtime_file(self):
        compose_data = read_compose_file(self._compose_template_file)
        for service, service_config in self._compose_service_configs.items():
            envs: List[Env] = []
            env_files = service_config.get_env_files()
            for env_file in env_files:
                envs += env_file.get_envs()
            envs += service_config.get_envs()
            compose_data = self._add_service_env(compose_data, service, envs)
        write_compose_file(self._compose_runtime_file, compose_data)

    def _add_service_env(
        self, compose_data: Any, service: str, envs: List[Env]
    ) -> Any:
        # service not found
        if 'services' not in compose_data or service not in compose_data['services']: # noqa
            self.log_error(f'Cannot find services.{service}')
            return compose_data
        # service has no environment definition
        if 'environment' not in compose_data['services'][service]:
            compose_data['services'][service]['environment'] = {
                env.name: self._get_env_compose_value(env) for env in envs
            }
            return compose_data
        # service environment is a map
        if isinstance(compose_data['services'][service]['environment'], dict):
            new_env_map = self._get_service_new_env_map(
                compose_data['services'][service]['environment'], envs
            )
            for key, value in new_env_map.items():
                compose_data['services'][service]['environment'][key] = value
            return compose_data
        # service environment is a list
        if isinstance(compose_data['services'][service]['environment'], list):
            new_env_list = self._get_service_new_env_list(
                compose_data['services'][service]['environment'], envs
            )
            compose_data['services'][service]['environment'] += new_env_list
            return compose_data
        return compose_data

    def _get_service_new_env_map(
        self, service_env_map: Mapping[str, str], new_envs: List[Env]
    ) -> Mapping[str, str]:
        new_service_envs: Mapping[str, str] = {}
        for env in new_envs:
            if env.name in service_env_map:
                continue
            new_service_envs[env.name] = self._get_env_compose_value(env)
        return new_service_envs

    def _get_service_new_env_list(
        self, service_env_list: List[str], new_envs: List[Env]
    ) -> List[str]:
        new_service_envs: List[str] = []
        for env in new_envs:
            should_be_added = 0 == len([
                service_env for service_env in service_env_list
                if service_env.startswith(env.name + '=')
            ])
            if not should_be_added:
                continue
            new_service_envs.append(
                env.name + '=' + self._get_env_compose_value(env)
            )
        return new_service_envs

    def _get_env_compose_value(self, env: Env) -> str:
        return '${' + env.name + ':-' + env.default + '}'

    def _add_compose_envs(self):
        data = read_compose_file(self._compose_template_file)
        env_map = fetch_compose_file_env_map(data)
        for key, value in env_map.items():
            # Need to get this everytime because we only want
            # the first compose file env value for a certain key
            existing_env_map = self._get_existing_env_map()
            if key in existing_env_map:
                continue
            os_name = key
            if self._compose_env_prefix != '':
                os_name = f'{self._compose_env_prefix}_{os_name}'
            self._envs.append(Env(name=key, os_name=os_name, default=value))

    def _get_existing_env_map(self) -> Mapping[str, str]:
        env_map: Mapping[str, str] = {}
        for env_file in self._env_files:
            envs = env_file.get_envs()
            env_map.update({
                env.name: env.default for env in envs
            })
        env_map.update({
            env.name: env.default for env in self._envs
        })
        return env_map

    def _get_compose_runtime_file(self, compose_file_name: str) -> str:
        directory, file = os.path.split(compose_file_name)
        prefix = '_' if file.startswith('.') else '._'
        runtime_prefix = self.get_cmd_name()
        if self._group is not None:
            group_prefix = to_cmd_name(self._group.get_complete_name())
            runtime_prefix = f'{group_prefix}-{runtime_prefix}'
        runtime_prefix += '-' + get_random_name(
            separator='-', add_random_digit=True, digit_count=3
        )
        runtime_prefix = '.' + runtime_prefix + '.runtime'
        file_parts = file.split('.')
        if len(file_parts) > 1:
            file_parts[-2] += runtime_prefix
            runtime_file_name = prefix + '.'.join(file_parts)
            return os.path.join(directory, runtime_file_name)
        runtime_file_name = prefix + file + runtime_prefix
        return os.path.join(directory, runtime_file_name)

    def _get_compose_template_file(self, compose_file: Optional[str]) -> str:
        if compose_file is None:
            for _compose_file in [
                'compose.yml', 'compose.yaml',
                'docker-compose.yml', 'docker-compose.yaml'
            ]:
                if os.path.exists(os.path.join(self.cwd, _compose_file)):
                    return os.path.join(self.cwd, _compose_file)
                    return
            raise Exception(f'Cannot find compose file on {self.cwd}')
        if os.path.isabs(compose_file) and os.path.exists(compose_file):
            return compose_file
        if os.path.exists(os.path.join(self.cwd, compose_file)):
            return os.path.join(self.cwd, compose_file)
        raise Exception(f'Invalid compose file: {compose_file}')

    def _get_cmd_str(self, *args: Any, **kwargs: Any) -> str:
        setup_cmd_str = self._create_cmd_str(
            self._setup_cmd_path, self._setup_cmd, *args, **kwargs
        )
        command_options = dict(self._compose_options)
        command_options['--file'] = self._compose_runtime_file
        options = ' '.join([
            f'{self.render_str(key)} {double_quote(self.render_str(val))}'
            for key, val in command_options.items()
            if self.render_str(val) != ''
        ])
        flags = ' '.join([
            self.render_str(flag) for flag in self._compose_flags
            if self.render_str(flag) != ''
        ])
        args = ' '.join([
            double_quote(self.render_str(arg)) for arg in self._compose_args
            if self.render_str(arg) != ''
        ])
        cmd_str = '\n'.join([
            setup_cmd_str,
            f'docker compose {options} {self._compose_cmd} {flags} {args}',
        ])
        self.log_info(f'Command: {cmd_str}')
        return cmd_str

    def __repr__(self) -> str:
        return f'<DockerComposeTask name={self._name}>'



from beartype.typing import (
    Any, Iterable, Callable, List, Mapping, Optional, TypeVar, Type, Tuple,
    Union
)

assert Any
assert Iterable
assert Callable
assert List
assert Mapping
assert Optional
assert TypeVar
assert Type
assert Tuple
assert Union



from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union
)
from zrb.helper.typecheck import typechecked
from zrb.config.config import show_time
from zrb.task.any_task import AnyTask
from zrb.helper.string.conversion import to_boolean, to_cmd_name
from zrb.helper.string.jinja import is_probably_jinja
from zrb.helper.render_data import DEFAULT_RENDER_DATA
from zrb.helper.log import logger
from zrb.helper.accessories.color import colored, get_random_color
from zrb.helper.accessories.icon import get_random_icon
from zrb.helper.util import coalesce_str
from zrb.task_input.any_input import AnyInput
from zrb.task_group.group import Group
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile

import asyncio
import datetime
import os
import time
import jinja2
import sys

LOG_NAME_LENGTH = 20


@typechecked
class CommonTaskModel():
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        description: str = '',
        inputs: List[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0,
        run: Optional[Callable[..., Any]] = None,
        skip_execution: Union[bool, str, Callable[..., bool]] = False
    ):
        self._name = name
        self._group = group
        if group is not None:
            group.add_task(self)
        self._description = coalesce_str(description, name)
        self._inputs = inputs
        self._envs = envs
        self._env_files = env_files
        self._icon = coalesce_str(icon, get_random_icon())
        self._color = coalesce_str(color, get_random_color())
        self._retry = retry
        self._retry_interval = retry_interval
        self._upstreams = upstreams
        self._checkers = checkers
        self._checking_interval = checking_interval
        self._run_function: Optional[Callable[..., Any]] = run
        self._skip_execution = skip_execution
        self._allow_add_envs = True
        self._allow_add_env_files = True
        self._allow_add_inputs = True

    def set_name(self, new_name: str):
        if self._description == self._name:
            self._description = new_name
        self._name = new_name

    def set_description(self, new_description: str):
        self._description = new_description

    def set_icon(self, new_icon: str):
        self._icon = new_icon

    def set_color(self, new_color: str):
        self._color = new_color

    def set_retry(self, new_retry: int):
        self._retry = new_retry

    def set_skip_execution(
        self, skip_execution: Union[bool, str, Callable[..., bool]]
    ):
        self._skip_execution = skip_execution

    def set_retry_interval(self, new_retry_interval: Union[float, int]):
        self._retry_interval = new_retry_interval

    def set_checking_interval(self, new_checking_interval: Union[float, int]):
        self._checking_interval = new_checking_interval

    def add_inputs(self, *inputs: AnyInput):
        if not self._allow_add_inputs:
            raise Exception(f'Cannot add inputs on `{self._name}`')
        self._inputs += inputs

    def add_envs(self, *envs: Env):
        if not self._allow_add_envs:
            raise Exception(f'Cannot add envs on `{self._name}`')
        self._envs += envs

    def add_env_files(self, *env_files: EnvFile):
        if not self._allow_add_env_files:
            raise Exception(f'Cannot add env_files on `{self._name}`')
        self._env_files += env_files

    def get_icon(self) -> str:
        return self._icon

    def get_color(self) -> str:
        return self._color

    def get_env_files(self) -> List[EnvFile]:
        return self._env_files

    def get_envs(self) -> List[Env]:
        return self._envs

    def get_inputs(self) -> List[AnyInput]:
        return self._inputs

    def get_checkers(self) -> Iterable[AnyTask]:
        return self._checkers

    def get_upstreams(self) -> Iterable[AnyTask]:
        return self._upstreams

    def get_description(self) -> str:
        return self._description

    def get_cmd_name(self) -> str:
        return to_cmd_name(self._name)


@typechecked
class TimeTracker():

    def __init__(self):
        self.__start_time: float = 0
        self.__end_time: float = 0

    def _start_timer(self):
        self.__start_time = time.time()

    def _end_timer(self):
        self.__end_time = time.time()

    def _get_elapsed_time(self) -> float:
        return self.__end_time - self.__start_time


@typechecked
class AttemptTracker():

    def __init__(self, retry: int = 2):
        self.__retry = retry
        self.__attempt: int = 1
        self.__no_more_attempt: bool = False

    def _get_max_attempt(self) -> int:
        return self.__retry + 1

    def _get_attempt(self) -> int:
        return self.__attempt

    def _increase_attempt(self):
        self.__attempt += 1

    def _should_attempt(self) -> bool:
        attempt = self._get_attempt()
        max_attempt = self._get_max_attempt()
        return attempt <= max_attempt

    def _is_last_attempt(self) -> bool:
        attempt = self._get_attempt()
        max_attempt = self._get_max_attempt()
        return attempt >= max_attempt


@typechecked
class FinishTracker():

    def __init__(self):
        self.__execution_queue: Optional[asyncio.Queue] = None
        self.__counter = 0

    async def _mark_awaited(self):
        if self.__execution_queue is None:
            self.__execution_queue = asyncio.Queue()
        self.__counter += 1

    async def _mark_done(self):
        # Tracker might be started several times
        # However, when the execution is marked as done, it applied globally
        # Thus, we need to send event as much as the counter.
        for i in range(self.__counter):
            await self.__execution_queue.put(True)

    async def _is_done(self) -> bool:
        while self.__execution_queue is None:
            await asyncio.sleep(0.05)
        return await self.__execution_queue.get()


@typechecked
class PidModel():

    def __init__(self):
        self.__task_pid: int = os.getpid()

    def _set_task_pid(self, pid: int):
        self.__task_pid = pid

    def _get_task_pid(self) -> int:
        return self.__task_pid


class AnyExtensionFileSystemLoader(jinja2.FileSystemLoader):
    def get_source(self, environment, template):
        for search_dir in self.searchpath:
            file_path = os.path.join(search_dir, template)
            if os.path.exists(file_path):
                with open(file_path, 'r') as file:
                    contents = file.read()
                return contents, file_path, lambda: False
        raise jinja2.TemplateNotFound(template)


@typechecked
class Renderer():

    def __init__(self):
        self.__input_map: Mapping[str, Any] = {}
        self.__env_map: Mapping[str, str] = {}
        self.__render_data: Optional[Mapping[str, Any]] = None
        self.__rendered_str: Mapping[str, str] = {}

    def get_input_map(self) -> Mapping[str, Any]:
        # This return reference to input map, so input map can be updated
        return self.__input_map

    def _set_input_map(self, key: str, val: Any):
        self.__input_map[key] = val

    def get_env_map(self) -> Mapping[str, str]:
        # This return reference to env map, so env map can be updated
        return self.__env_map

    def _set_env_map(self, key: str, val: str):
        self.__env_map[key] = val

    def render_any(
        self, val: Any, data: Optional[Mapping[str, Any]] = None
    ) -> Any:
        if isinstance(val, str):
            return self.render_str(val, data)
        return val

    def render_float(
        self, val: Union[str, float], data: Optional[Mapping[str, Any]] = None
    ) -> float:
        if isinstance(val, str):
            return float(self.render_str(val, data))
        return val

    def render_int(
        self, val: Union[str, int], data: Optional[Mapping[str, Any]] = None
    ) -> int:
        if isinstance(val, str):
            return int(self.render_str(val, data))
        return val

    def render_bool(
        self, val: Union[str, bool], data: Optional[Mapping[str, Any]] = None
    ) -> bool:
        if isinstance(val, str):
            return to_boolean(self.render_str(val, data))
        return val

    def render_str(
        self, val: str, data: Optional[Mapping[str, Any]] = None
    ) -> str:
        if val in self.__rendered_str:
            return self.__rendered_str[val]
        if not is_probably_jinja(val):
            return val
        template = jinja2.Template(val)
        render_data = self._get_render_data(additional_data=data)
        try:
            rendered_text = template.render(render_data)
        except Exception:
            raise Exception(f'Fail to render "{val}" with data: {render_data}')
        self.__rendered_str[val] = rendered_text
        return rendered_text

    def render_file(
        self, location: str, data: Optional[Mapping[str, Any]] = None
    ) -> str:
        location_dir = os.path.dirname(location)
        env = jinja2.Environment(
            loader=AnyExtensionFileSystemLoader([location_dir])
        )
        template = env.get_template(location)
        render_data = self._get_render_data(additional_data=data)
        render_data['TEMPLATE_DIR'] = location_dir
        rendered_text = template.render(render_data)
        return rendered_text

    def _get_render_data(
        self, additional_data: Optional[Mapping[str, Any]] = None
    ) -> Mapping[str, Any]:
        self._ensure_cached_render_data()
        if additional_data is None:
            return self.__render_data
        return {**self.__render_data, **additional_data}

    def _ensure_cached_render_data(self):
        if self.__render_data is not None:
            return self.__render_data
        render_data = dict(DEFAULT_RENDER_DATA)
        render_data.update({
            'env': self.__env_map,
            'input': self.__input_map,
        })
        self.__render_data = render_data
        return render_data


@typechecked
class TaskModelWithPrinterAndTracker(
    CommonTaskModel, PidModel, TimeTracker
):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        description: str = '',
        inputs: List[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
        retry_interval: Union[int, float] = 1,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[int, float] = 0,
        run: Optional[Callable[..., Any]] = None,
        skip_execution: Union[bool, str, Callable[..., bool]] = False

    ):
        self._filled_complete_name: Optional[str] = None
        self._has_cli_interface = False
        self._complete_name: Optional[str] = None
        CommonTaskModel.__init__(
            self,
            name=name,
            group=group,
            description=description,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            retry=retry,
            retry_interval=retry_interval,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            run=run,
            skip_execution=skip_execution,
        )
        PidModel.__init__(self)
        TimeTracker.__init__(self)

    def log_debug(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.debug(colored_message)

    def log_warn(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.warning(colored_message)

    def log_info(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.info(colored_message)

    def log_error(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.error(colored_message, exc_info=True)

    def log_critical(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.critical(colored_message, exc_info=True)

    def print_out(self, message: Any, trim_message: bool = True):
        prefix = self._get_colored_print_prefix()
        message_str = f'{message}'.rstrip() if trim_message else f'{message}'
        print(f'🤖 ○ {prefix} • {message_str}', file=sys.stderr)
        sys.stderr.flush()

    def print_err(self, message: Any, trim_message: bool = True):
        prefix = self._get_colored_print_prefix()
        message_str = f'{message}'.rstrip() if trim_message else f'{message}'
        print(f'🤖 △ {prefix} • {message_str}', file=sys.stderr)
        sys.stderr.flush()

    def print_out_dark(self, message: Any, trim_message: bool = True):
        message_str = f'{message}'
        self.print_out(colored(message_str, attrs=['dark']), trim_message)

    def _play_bell(self):
        print('\a', end='', file=sys.stderr)

    def _get_colored_print_prefix(self) -> str:
        return self._get_colored(self._get_print_prefix())

    def _get_colored(self, text: str) -> str:
        return colored(text, color=self.get_color())

    def _get_print_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=show_time)
        icon = self.get_icon()
        truncated_name = self._get_filled_complete_name()
        return f'{common_prefix} {icon} {truncated_name}'

    def _get_log_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=False)
        icon = self.get_icon()
        filled_name = self._get_filled_complete_name()
        return f'{common_prefix} {icon} {filled_name}'

    def _get_common_prefix(self, show_time: bool) -> str:
        attempt = self._get_attempt()
        max_attempt = self._get_max_attempt()
        pid = self._get_task_pid()
        if show_time:
            now = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
            return f'◷ {now} ❁ {pid} → {attempt}/{max_attempt}'
        return f'❁ {pid} → {attempt}/{max_attempt}'

    def _get_filled_complete_name(self) -> str:
        if self._filled_complete_name is not None:
            return self._filled_complete_name
        complete_name = self._get_complete_name()
        self._filled_complete_name = complete_name.rjust(LOG_NAME_LENGTH, ' ')
        return self._filled_complete_name

    def _get_complete_name(self) -> str:
        if self._complete_name is not None:
            return self._complete_name
        executable_prefix = ''
        if self._has_cli_interface:
            executable_prefix += self._get_executable_name() + ' '
        cmd_name = self.get_cmd_name()
        if self._group is None:
            self._complete_name = f'{executable_prefix}{cmd_name}'
            return self._complete_name
        group_cmd_name = self._group.get_complete_name()
        self._complete_name = f'{executable_prefix}{group_cmd_name} {cmd_name}'
        return self._complete_name

    def _get_executable_name(self) -> str:
        if len(sys.argv) > 0 and sys.argv[0] != '':
            return os.path.basename(sys.argv[0])
        return 'zrb'

    def set_has_cli_interface(self):
        self._has_cli_interface = True



from typing import List
from core.schema import BaseDateTimeSchema, BaseCountSchema


class PascalZrbEntityNameData(BaseDateTimeSchema):
    snake_zrb_column_name: str


class PascalZrbEntityName(PascalZrbEntityNameData):
    id: str

    class Config:
        orm_mode = True


class PascalZrbEntityNameResult(BaseCountSchema):
    data: List[PascalZrbEntityName]



from typing import Any, Callable
from abc import ABC, abstractmethod
import jsons


class Serializer(ABC):
    @abstractmethod
    def encode(self, message: Any) -> Any:
        pass

    @abstractmethod
    def decode(self, message: Any) -> Any:
        pass


class CustomSerializer(Serializer):
    def __init__(
        self, encoder: Callable[[Any], Any], decoder: Callable[[Any], Any]
    ):
        self.encoder = encoder
        self.decoder = decoder

    def encode(self, message: Any) -> Any:
        return self.encoder(message)

    def decode(self, encoded_message: Any) -> Any:
        return self.decoder(encoded_message)


class JsonSerializer(Serializer):
    def encode(self, message: Any) -> Any:
        return jsons.dumps(message).encode()

    def decode(self, encoded_message: Any) -> Any:
        return jsons.loads(encoded_message.decode())



from zrb.helper.typing import Any, Callable, Iterable, Optional, Union, TypeVar
from zrb.helper.typecheck import typechecked
from zrb.task.base_task import BaseTask
from zrb.task.any_task import AnyTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

import socket
import asyncio

TPortChecker = TypeVar('TPortChecker', bound='PortChecker')


@typechecked
class PortChecker(BaseTask):

    def __init__(
        self,
        name: str = 'port-check',
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        host: str = 'localhost',
        port: Union[int, str] = 80,
        timeout: Union[int, str] = 5,
        upstreams: Iterable[AnyTask] = [],
        checking_interval: float = 0.1,
        show_error_interval: float = 5,
        skip_execution: Union[bool, str, Callable[..., bool]] = False
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=checking_interval,
            retry=0,
            retry_interval=0,
            skip_execution=skip_execution,
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self._show_error_interval = show_error_interval

    def copy(self) -> TPortChecker:
        return super().copy()

    def to_function(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., bool]:
        return super().to_function(env_prefix, raise_error)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        host = self.render_str(self._host)
        port = self.render_int(self._port)
        timeout = self.render_int(self._timeout)
        wait_time = 0
        while not self._check_port(
            host=host,
            port=port,
            timeout=timeout,
            should_print_error=wait_time >= self._show_error_interval
        ):
            if wait_time >= self._show_error_interval:
                wait_time = 0
            await asyncio.sleep(self._checking_interval)
            wait_time += self._checking_interval
        return True

    def _check_port(
        self, host: str, port: int, timeout: int, should_print_error: bool
    ) -> bool:
        label = self._get_label(host, port)
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(timeout)
                result = sock.connect_ex((host, port))
                if result == 0:
                    self.print_out(f'{label} (OK)')
                    return True
                self._debug_and_print_error(
                    f'{label} (Not OK)', should_print_error
                )
        except Exception:
            self._debug_and_print_error(
                f'{label} Socket error', should_print_error
            )
        return False

    def _debug_and_print_error(self, message: str, should_print_error: bool):
        if should_print_error:
            self.print_err(message)
        self.log_debug(message)

    def _get_label(self, host: str, port: int) -> str:
        return f'Checking {host}:{port}'

    def __repr__(self) -> str:
        return f'<PortChecker name={self._name}>'



from zrb.builtin.group import project_group
from zrb import Task, runner

destroy_project = Task(
    name='destroy',
    group=project_group,
    upstreams=[],
    description='Remove project deployment',
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
)
runner.register(destroy_project)



from sqlalchemy import Column, String
from core.repo import Repo, DBEntityMixin, DBRepo
from module.snake_zrb_module_name.schema.snake_zrb_entity_name import (
    PascalZrbEntityName, PascalZrbEntityNameData
)
from module.snake_zrb_module_name.component import Base


class DBEntityPascalZrbEntityName(Base, DBEntityMixin):
    __tablename__ = "snake_zrb_plural_entity_name"
    snake_zrb_column_name = Column(String)


class PascalZrbEntityNameRepo(Repo[PascalZrbEntityName, PascalZrbEntityNameData]):
    pass


class PascalZrbEntityNameDBRepo(
    DBRepo[PascalZrbEntityName, PascalZrbEntityNameData], PascalZrbEntityNameRepo
):
    schema_cls = PascalZrbEntityName
    db_entity_cls = DBEntityPascalZrbEntityName



from core.model import RepoModel
from module.log.schema.activity import (
    Activity, ActivityData, ActivityResult
)


class ActivityModel(
    RepoModel[Activity, ActivityData, ActivityResult]
):
    schema_result_cls = ActivityResult



from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from config import (
    zrb_app_name, app_enable_event_handler, app_enable_rpc_server,
    app_enable_frontend, app_db_auto_migrate
)
from config import app_src_dir
from migrate import migrate
from component.app_state import app_state, set_not_ready_on_error
from component.messagebus import consumer
from component.rpc import rpc_server
from component.log import logger
from helper.async_task import create_task
from contextlib import asynccontextmanager
import os


@asynccontextmanager
async def app_lifespan(app: FastAPI):
    logger.info(f'{zrb_app_name} started')
    if app_db_auto_migrate:
        await migrate()
    app_state.set_liveness(True)
    if app_enable_event_handler:
        create_task(consumer.start(), on_error=set_not_ready_on_error)
    if app_enable_rpc_server:
        create_task(rpc_server.start(), on_error=set_not_ready_on_error)
    if app_enable_frontend:
        build_path = os.path.join(app_src_dir, 'frontend', 'build')
        app.mount(
            path='',
            app=StaticFiles(directory=build_path, html=True),
            name='frontend-static-resources'
        )
    app_state.set_readiness(True)
    logger.info(f'{zrb_app_name} started')
    yield
    if app_enable_event_handler:
        await consumer.stop()
    logger.info(f'{zrb_app_name} closed')



from zrb.builtin.group import project_group
from zrb import Task, runner

remove_project_containers = Task(
    name='remove-containers',
    group=project_group,
    upstreams=[],
    description='Remove project containers',
    run=lambda *args, **kwargs: kwargs.get('_task').print_out('🆗')
)
runner.register(remove_project_containers)



from module.log.core.historical_repo_model import HistoricalRepoModel
from module.snake_zrb_module_name.schema.snake_zrb_entity_name import (
    PascalZrbEntityName, PascalZrbEntityNameData, PascalZrbEntityNameResult
)


class PascalZrbEntityNameModel(
    HistoricalRepoModel[PascalZrbEntityName, PascalZrbEntityNameData, PascalZrbEntityNameResult]
):
    schema_result_cls = PascalZrbEntityNameResult
    log_entity_name = 'snake_zrb_entity_name'



from zrb.helper.typecheck import typechecked
import random


@typechecked
def get_random_icon() -> str:
    icons = [
        '🐶', '🐱', '🐭', '🐹', '🦊', '🐻', '🐨', '🐯', '🦁', '🐮', '🐷', '🍎', '🍐',
        '🍊', '🍋', '🍌', '🍉', '🍇', '🍓', '🍈', '🍒', '🍑'
    ]
    return random.choice(icons)



from zrb.builtin.explain import (
    explain_dry_principle, explain_kiss_principle, explain_yagni_principle,
    explain_solid_principle, explain_zen_of_python
)


def test_explain_dry():
    function = explain_dry_principle.to_function()
    result = function()
    assert result is None


def test_explain_kiss():
    function = explain_kiss_principle.to_function()
    result = function()
    assert result is None


def test_explain_solid():
    function = explain_solid_principle.to_function()
    result = function()
    assert result is None


def test_explain_yagni():
    function = explain_yagni_principle.to_function()
    result = function()
    assert result is None


def test_explain_zen_of_python():
    function = explain_zen_of_python.to_function()
    result = function()
    assert result is None



from zrb.helper.typing import (
    Any, Callable, Iterable, List, Mapping, Optional, Union, TypeVar
)
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task.base_task import BaseTask
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput
from zrb.config.config import default_shell

import asyncio
import atexit
import os
import pathlib
import signal
import sys
import subprocess
import time

_has_stty = True
try:
    _original_stty = subprocess.getoutput('stty -g').rstrip()
except Exception:
    _has_stty = False


def _reset_stty():
    if _has_stty:
        subprocess.run(['stty', _original_stty])


CmdVal = Union[str, Iterable[str], Callable[..., Union[Iterable[str], str]]]
TCmdTask = TypeVar('TCmdTask', bound='CmdTask')


class CmdResult():
    def __init__(self, output: str, error: str):
        self.output = output
        self.error = error


class CmdGlobalState():
    def __init__(self):
        self.no_more_attempt: bool = False
        self.is_killed_by_signal: bool = False


@typechecked
class CmdTask(BaseTask):
    '''
    Command Task.
    You can use this task to run shell command.

    For example:
    ```python
    # run a simple task
    hello = CmdTask(
        name='hello',
        inputs=[StrInput(name='name', default='World')],
        envs=[Env(name='HOME_DIR', os_name='HOME')],
        cmd=[
            'echo Hello {{ input.name }}',
            'echo Home directory is: $HOME_DIR',
        ]
    )
    runner.register(hello)

    # run a long running process
    run_server = CmdTask(
        name='run',
        inputs=[StrInput(name='dir', default='.')],
        envs=[Env(name='PORT', os_name='WEB_PORT', default='3000')],
        cmd='python -m http.server $PORT --directory {{input.dir}}',
        checkers=[HTTPChecker(port='{{env.PORT}}')]
    )
    runner.register(run_server)
    ```
    '''

    _pids: List[int] = []
    _global_state = CmdGlobalState()

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        executable: Optional[str] = None,
        cmd: CmdVal = '',
        cmd_path: CmdVal = '',
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[AnyTask] = [],
        checkers: Iterable[AnyTask] = [],
        checking_interval: Union[float, int] = 0,
        retry: int = 2,
        retry_interval: Union[float, int] = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        skip_execution: Union[bool, str, Callable[..., bool]] = False,
        return_upstream_result: bool = False
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            skip_execution=skip_execution,
            return_upstream_result=return_upstream_result
        )
        max_output_line = max_output_line if max_output_line > 0 else 1
        max_error_line = max_error_line if max_error_line > 0 else 1
        self._cmd = cmd
        self._cmd_path = cmd_path
        self._set_cwd(cwd)
        self._max_output_size = max_output_line
        self._max_error_size = max_error_line
        self._output_buffer: Iterable[str] = []
        self._error_buffer: Iterable[str] = []
        if executable is None and default_shell != '':
            executable = default_shell
        self._executable = executable
        self._process: Optional[asyncio.subprocess.Process]
        self._preexec_fn = preexec_fn

    def copy(self) -> TCmdTask:
        return super().copy()

    def _set_cwd(
        self, cwd: Optional[Union[str, pathlib.Path]]
    ):
        if cwd is None:
            self.cwd: Union[str, pathlib.Path] = os.getcwd()
            return
        self.cwd: Union[str, pathlib.Path] = cwd

    def to_function(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., CmdResult]:
        return super().to_function(env_prefix, raise_error)

    def print_result(self, result: CmdResult):
        if result.output == '':
            return
        print(result.output)

    def _get_shell_env_map(self) -> Mapping[str, Any]:
        env_map = self.get_env_map()
        input_map = self.get_input_map()
        for input_name, input_value in input_map.items():
            upper_input_name = '_INPUT_' + input_name.upper()
            if upper_input_name not in env_map:
                env_map[upper_input_name] = f'{input_value}'
        return env_map

    async def run(self, *args: Any, **kwargs: Any) -> CmdResult:
        cmd = self._get_cmd_str(*args, **kwargs)
        env_map = self._get_shell_env_map()
        self.print_out_dark('Run script: ' + self._get_multiline_repr(cmd))
        self.print_out_dark('Working directory: ' + self.cwd)
        self._output_buffer = []
        self._error_buffer = []
        process = await asyncio.create_subprocess_shell(
            cmd,
            cwd=self.cwd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=env_map,
            shell=True,
            executable=self._executable,
            close_fds=True,
            preexec_fn=self._preexec_fn,
            bufsize=0
        )
        self._set_task_pid(process.pid)
        self._pids.append(process.pid)
        self._process = process
        signal.signal(signal.SIGINT, self._on_kill)
        signal.signal(signal.SIGTERM, self._on_kill)
        atexit.register(self._on_exit)
        await self._wait_process(process)
        self.log_info('Process completed')
        atexit.unregister(self._on_exit)
        output = '\n'.join(self._output_buffer)
        error = '\n'.join(self._error_buffer)
        # get return code
        return_code = process.returncode
        self.log_info(f'Exit status: {return_code}')
        if return_code != 0 and not self._global_state.is_killed_by_signal:
            raise Exception(
                f'Process {self._name} exited ({return_code}): {error}'
            )
        return CmdResult(output, error)

    def _should_attempt(self) -> bool:
        if self._global_state.no_more_attempt:
            return False
        return super()._should_attempt()

    def _is_last_attempt(self) -> bool:
        if self._global_state.no_more_attempt:
            return True
        return super()._is_last_attempt()

    def _on_kill(self, signum: Any, frame: Any):
        self._global_state.no_more_attempt = True
        self._global_state.is_killed_by_signal = True
        self.print_out_dark(f'Getting signal {signum}')
        for pid in self._pids:
            self._kill_by_pid(pid)
        self.print_out_dark(f'Exiting with signal {signum}')
        sys.exit(signum)

    def _on_exit(self):
        self._global_state.no_more_attempt = True
        self._kill_by_pid(self._process.pid)

    def _kill_by_pid(self, pid: int):
        '''
        Kill a pid, gracefully
        '''
        try:
            process_ever_exists = False
            if self._is_process_exist(pid):
                process_ever_exists = True
                self.print_out_dark(f'Send SIGTERM to process {pid}')
                os.killpg(os.getpgid(pid), signal.SIGTERM)
                time.sleep(0.5)
            if self._is_process_exist(pid):
                self.print_out_dark(f'Send SIGINT to process {pid}')
                os.killpg(os.getpgid(pid), signal.SIGINT)
                time.sleep(0.5)
            if self._is_process_exist(pid):
                self.print_out_dark(f'Send SIGKILL to process {pid}')
                os.killpg(os.getpgid(pid), signal.SIGKILL)
            if process_ever_exists:
                self.print_out_dark(f'Process {pid} is killed successfully')
        except Exception:
            self.log_error(f'Cannot kill process {pid}')

    def _is_process_exist(self, pid: int) -> bool:
        try:
            os.killpg(os.getpgid(pid), 0)
            return True
        except ProcessLookupError:
            return False

    async def _wait_process(self, process: asyncio.subprocess.Process):
        # Create queue
        stdout_queue = asyncio.Queue()
        stderr_queue = asyncio.Queue()
        # Read from streams and put into queue
        stdout_process = asyncio.create_task(self._queue_stream(
            process.stdout, stdout_queue
        ))
        stderr_process = asyncio.create_task(self._queue_stream(
            process.stderr, stderr_queue
        ))
        # Handle messages in queue
        stdout_log_process = asyncio.create_task(self._log_from_queue(
            stdout_queue, self.print_out,
            self._output_buffer, self._max_output_size
        ))
        stderr_log_process = asyncio.create_task(self._log_from_queue(
            stderr_queue, self.print_err,
            self._error_buffer, self._max_error_size
        ))
        # wait process
        await process.wait()
        # wait reader and logger
        await stdout_process
        await stderr_process
        await stdout_queue.put(None)
        await stderr_queue.put(None)
        await stdout_log_process
        await stderr_log_process

    def _get_cmd_str(self, *args: Any, **kwargs: Any) -> str:
        return self._create_cmd_str(self._cmd_path, self._cmd, *args, **kwargs)

    def _create_cmd_str(
        self, cmd_path: CmdVal, cmd: CmdVal, *args: Any, **kwargs: Any
    ) -> str:
        if not isinstance(cmd_path, str) or cmd_path != '':
            if callable(cmd_path):
                return self._render_cmd_path_str(cmd_path(*args, **kwargs))
            return self._render_cmd_path_str(cmd_path)
        if callable(cmd):
            return self._render_cmd_str(cmd(*args, **kwargs))
        return self._render_cmd_str(cmd)

    def _render_cmd_path_str(self, cmd_path: Union[str, Iterable[str]]) -> str:
        if isinstance(cmd_path, str):
            return self.render_file(cmd_path)
        return '\n'.join([
            self.render_file(cmd_path_str)
            for cmd_path_str in cmd_path
        ])

    def _render_cmd_str(self, cmd: Union[str, Iterable[str]]) -> str:
        if isinstance(cmd, str):
            return self.render_str(cmd)
        return self.render_str('\n'.join(list(cmd)))

    async def _queue_stream(self, stream, queue: asyncio.Queue):
        while True:
            try:
                line = await stream.readline()
            except Exception as e:
                line = str(e)
            if not line:
                break
            await queue.put(line)

    async def _log_from_queue(
        self,
        queue: asyncio.Queue,
        print_log: Callable[[str], None],
        buffer: Iterable[str],
        max_size: int
    ):
        while True:
            line = await queue.get()
            if not line:
                break
            line_str = line.decode('utf-8').rstrip()
            self._add_to_buffer(buffer, max_size, line_str)
            _reset_stty()
            print_log(line_str)
            _reset_stty()

    def _add_to_buffer(
        self, buffer: Iterable[str], max_size: int, new_line: str
    ):
        if len(buffer) >= max_size:
            buffer.pop(0)
        buffer.append(new_line)
    
    def __repr__(self) -> str:
        return f'<CmdTask name={self._name}>'



from zrb.helper.typing import Any
from zrb.builtin.generator.common.helper import (
    validate_existing_project_dir, validate_inexisting_automation
)
from zrb.builtin.generator.common.task_input import (
    project_dir_input, app_name_input, app_image_name_input, http_port_input,
    env_prefix_input
)
from zrb.builtin.generator.common.task_factory import create_register_module
from zrb.builtin.generator.project_task.task_factory import (
    create_ensure_project_tasks, create_add_build_images_upstream,
    create_add_deploy_upstream, create_add_destroy_upstream,
    create_add_push_images_upstream, create_add_remove_containers_upstream,
    create_add_start_containers_upstream, create_add_start_upstream,
    create_add_stop_containers_upstream
)
from zrb.builtin.group import project_add_group
from zrb.task.task import Task
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker
from zrb.runner import runner
from zrb.helper.util import to_kebab_case

import os

CURRENT_DIR = os.path.dirname(__file__)
SNAKE_APP_NAME_TPL = '{{util.to_snake_case(input.app_name)}}'

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name='validate',
    inputs=[project_dir_input, app_name_input],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_existing_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    validate_inexisting_automation(project_dir, app_name)
    source_dir = os.path.join(project_dir, 'src', f'{to_kebab_case(app_name)}')
    if os.path.exists(source_dir):
        raise Exception(f'Source already exists: {source_dir}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=[
        project_dir_input,
        app_name_input,
        app_image_name_input,
        http_port_input,
        env_prefix_input,
    ],
    upstreams=[validate],
    replacements={
        'zrbAppName': '{{input.app_name}}',
        'zrbAppHttpPort': '{{util.coalesce(input.http_port, "3000")}}',
        'ZRB_ENV_PREFIX': '{{util.coalesce(input.env_prefix, "MY").upper()}}',
        'zrb-app-image-name': '{{input.app_image_name}}'
    },
    template_path=os.path.join(CURRENT_DIR, 'template'),
    destination_path='{{ input.project_dir }}',
    excludes=[
        '*/deployment/venv',
        '*/__pycache__',
    ]
)

register_local_module = create_register_module(
    module_path=f'_automate.{SNAKE_APP_NAME_TPL}.local',
    alias=f'{SNAKE_APP_NAME_TPL}_local',
    inputs=[app_name_input],
    upstreams=[copy_resource]
)

register_container_module = create_register_module(
    module_path=f'_automate.{SNAKE_APP_NAME_TPL}.container',
    alias=f'{SNAKE_APP_NAME_TPL}_container',
    inputs=[app_name_input],
    upstreams=[register_local_module]
)

register_image_module = create_register_module(
    module_path=f'_automate.{SNAKE_APP_NAME_TPL}.image',
    alias=f'{SNAKE_APP_NAME_TPL}_image',
    inputs=[app_name_input],
    upstreams=[register_container_module]
)

register_deployment_module = create_register_module(
    module_path=f'_automate.{SNAKE_APP_NAME_TPL}.deployment',
    alias=f'{SNAKE_APP_NAME_TPL}_deployment',
    inputs=[app_name_input],
    upstreams=[register_image_module]
)

ensure_project_tasks = create_ensure_project_tasks(
    upstreams=[copy_resource]
)

add_start_upstream = create_add_start_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.local',
    upstream_task_var=f'start_{SNAKE_APP_NAME_TPL}',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_start_container_upstream = create_add_start_containers_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.container',
    upstream_task_var=f'start_{SNAKE_APP_NAME_TPL}_container',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_stop_container_upstream = create_add_stop_containers_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.container',
    upstream_task_var=f'stop_{SNAKE_APP_NAME_TPL}_container',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_remove_container_upstream = create_add_remove_containers_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.container',
    upstream_task_var=f'remove_{SNAKE_APP_NAME_TPL}_container',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_build_image_upstream = create_add_build_images_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.image',
    upstream_task_var=f'build_{SNAKE_APP_NAME_TPL}_image',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_push_image_upstream = create_add_push_images_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.image',
    upstream_task_var=f'push_{SNAKE_APP_NAME_TPL}_image',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_deploy_upstream = create_add_deploy_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.deployment',
    upstream_task_var=f'deploy_{SNAKE_APP_NAME_TPL}',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)

add_destroy_upstream = create_add_destroy_upstream(
    upstream_module=f'_automate.{SNAKE_APP_NAME_TPL}.deployment',
    upstream_task_var=f'destroy_{SNAKE_APP_NAME_TPL}',
    upstreams=[ensure_project_tasks],
    inputs=[app_name_input]
)


@python_task(
    name='simple-python-app',
    group=project_add_group,
    upstreams=[
        register_local_module,
        register_container_module,
        register_image_module,
        register_deployment_module,
        add_start_upstream,
        add_start_container_upstream,
        add_stop_container_upstream,
        add_remove_container_upstream,
        add_build_image_upstream,
        add_push_image_upstream,
        add_deploy_upstream,
        add_destroy_upstream
    ],
    runner=runner
)
async def add_simple_python_app(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from zrb.task.decorator import python_task


def test_task_with_decorated_runner():
    @python_task(
        name='hello'
    )
    async def hello(*args, **kwargs) -> str:
        return 'hello'

    function = hello.to_function()
    result = function()
    assert result == 'hello'



from zrb.helper.typing import (
    Any, Callable, Iterable, Optional, Union
)
from zrb.helper.typecheck import typechecked
from zrb.task.any_task import AnyTask
from zrb.task_input.any_input import AnyInput
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.action.runner import Runner
from zrb.task.task import Task


@typechecked
def python_task(
    name: str,
    group: Optional[Group] = None,
    inputs: Iterable[AnyInput] = [],
    envs: Iterable[Env] = [],
    env_files: Iterable[EnvFile] = [],
    icon: Optional[str] = None,
    color: Optional[str] = None,
    description: str = '',
    upstreams: Iterable[AnyTask] = [],
    checkers: Iterable[AnyTask] = [],
    checking_interval: float = 0.1,
    retry: int = 2,
    retry_interval: float = 1,
    skip_execution: Union[bool, str, Callable[..., bool]] = False,
    return_upstream_result: bool = False,
    runner: Optional[Runner] = None
) -> Callable[[Callable[..., Any]], Task]:
    def _create_task(fn: Callable[..., Any]) -> Task:
        task = Task(
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            run=fn,
            skip_execution=skip_execution,
            return_upstream_result=return_upstream_result
        )
        if runner is not None:
            runner.register(task)
        return task
    return _create_task



from typing import Mapping, Any, List
from zrb import DockerComposeTask, Env, ServiceConfig, EnvFile, runner, Task
from zrb.helper.util import to_kebab_case
from zrb.builtin.group import project_group
from ._common import (
    APP_TEMPLATE_ENV_FILE_NAME, RESOURCE_DIR, MODULES, app_container_checker,
    rabbitmq_checker, rabbitmq_management_checker, redpanda_console_checker,
    kafka_outside_checker, kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker, local_input, run_mode_input,
    enable_monitoring_input, host_input, https_input, app_enable_otel_env
)
from .image import build_snake_zrb_app_name_image, image_input, image_env
import os

DOCKER_COMPOSE_APP_ENV_FILE_NAME = os.path.join(
    RESOURCE_DIR, 'docker-compose-app.env'
)

###############################################################################
# Functions
###############################################################################


def setup_runtime_compose_profile(*args: Any, **kwargs: Any) -> str:
    task: Task = kwargs.get('_task')
    env_map = task.get_env_map()
    compose_profiles: List[str] = [
        kwargs.get('snake_zrb_app_name_run_mode', 'monolith'),
    ]
    broker_type = env_map.get('APP_BROKER_TYPE', 'rabbitmq')
    if broker_type in ['rabbitmq', 'kafka']:
        compose_profiles.append(broker_type)
    if kwargs.get('enable_snake_zrb_app_name_monitoring', False):
        compose_profiles.append('monitoring')
    compose_profile_str = ','.join(compose_profiles)
    return f'export COMPOSE_PROFILES={compose_profile_str}'


def setup_all_compose_profile(*args: Any, **kwargs: Any) -> str:
    compose_profiles = [
        'monitoring',
        'monolith',
        'microservices',
        'kafka',
        'rabbitmq'
    ]
    compose_profile_str = ','.join(compose_profiles)
    return f'export COMPOSE_PROFILES={compose_profile_str}'


def skip_execution(*args: Any, **kwargs: Any) -> bool:
    return not kwargs.get('local_snake_zrb_app_name', True)


###############################################################################
# Env File Definitions
###############################################################################

compose_env_file = EnvFile(
    env_file=os.path.join(RESOURCE_DIR, 'docker-compose.env'),
    prefix='CONTAINER_ZRB_ENV_PREFIX'
)

###############################################################################
# Compose Service Config Definitions
###############################################################################

service_configs: Mapping[str, ServiceConfig] = {}
service_names = [to_kebab_case(module) + '-service' for module in MODULES]
for suffix in ['', 'gateway'] + service_names:
    service_suffix = '-' + suffix if suffix != '' else ''
    service_name = f'kebab-zrb-app-name{service_suffix}'
    service_configs[service_name] = ServiceConfig(
        env_files=[
            EnvFile(
                env_file=APP_TEMPLATE_ENV_FILE_NAME,
                prefix='CONTAINER_ZRB_ENV_PREFIX'
            ),
            EnvFile(
                env_file=DOCKER_COMPOSE_APP_ENV_FILE_NAME,
                prefix='CONTAINER_ZRB_ENV_PREFIX'
            )
        ],
        envs=[
            Env(
                name='APP_OTEL_EXPORTER_OTLP_ENDPOINT',
                os_name='',
                default='http://otel-collector:4317'
            )
        ]
    )

###############################################################################
# Task Definitions
###############################################################################

remove_snake_zrb_app_name_container = DockerComposeTask(
    icon='💨',
    name='remove-kebab-zrb-app-name-container',
    description='Remove human readable zrb app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    setup_cmd=setup_all_compose_profile,
    compose_cmd='down',
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=service_configs,
    envs=[image_env],
    env_files=[compose_env_file]
)
runner.register(remove_snake_zrb_app_name_container)

init_snake_zrb_app_name_container = DockerComposeTask(
    icon='🔥',
    name='init-kebab-zrb-app-name-container',
    group=project_group,
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        image_input,
    ],
    skip_execution=skip_execution,
    upstreams=[
        build_snake_zrb_app_name_image,
        remove_snake_zrb_app_name_container
    ],
    cwd=RESOURCE_DIR,
    setup_cmd=setup_runtime_compose_profile,
    compose_cmd='up',
    compose_flags=['-d'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=service_configs,
    envs=[
        image_env,
        app_enable_otel_env,
    ],
    env_files=[compose_env_file],
)

start_snake_zrb_app_name_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-zrb-app-name-container',
    description='Start human readable zrb app name container',
    group=project_group,
    inputs=[
        local_input,
        run_mode_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution=skip_execution,
    upstreams=[init_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    setup_cmd=setup_runtime_compose_profile,
    compose_cmd='logs',
    compose_flags=['-f'],
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=service_configs,
    envs=[image_env],
    env_files=[compose_env_file],
    checkers=[
        app_container_checker,
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)
runner.register(start_snake_zrb_app_name_container)

stop_snake_zrb_app_name_container = DockerComposeTask(
    icon='⛔',
    name='stop-kebab-zrb-app-name-container',
    description='Stop human readable zrb app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    setup_cmd=setup_all_compose_profile,
    compose_cmd='stop',
    compose_env_prefix='CONTAINER_ZRB_ENV_PREFIX',
    compose_service_configs=service_configs,
    envs=[image_env],
    env_files=[compose_env_file]
)
runner.register(stop_snake_zrb_app_name_container)
